/**
 * See readme for program information.
 *
 * How to run:
 * 
 * Compile:
 *   $ javac ClassViewer.java
 * 
 * Run (eyebot connected to /dev/ttyS0):
 *  Sets the channel in raw mode:
 *   $ stty -F /dev/ttyS0 speed 115200 raw crtscts
 *  Run the program:
 *   $ java ClassViewer < /dev/ttyS0
 * 
 * One-line version:
 *   $ javac ClassViewer.java && \
 *       stty -F /dev/ttyS0 speed 115200 raw crtscts && \
 *       java ClassViewer < /dev/ttyS0
 */

import java.awt.*;
import java.awt.geom.AffineTransform;
import java.awt.image.BufferedImage;
import java.io.*;
import java.util.Arrays;
import javax.swing.*;

/**
 * The main class.
 * 1. It sets up the frame and the component to render the image on.
 * 2. Creates a camreader with a stream and where to render the image on.
 */
class ClassViewer extends JFrame {
    ClassViewer() throws Exception {
	super("Eyebot classified image viewer 1.0");
	
	// create and add the component to render the image on
	CamComponent cc = new CamComponent();
	getContentPane().add(cc);

	// set up the frame
	setVisible(true);
	setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
	setSize(CamReader.IMG_WIDTH*3, CamReader.IMG_HEIGHT*3);
	
	// create the camreader
	CamReader cr = new CamReader();
	Graphics g = cc.bi.createGraphics();
	DataInputStream dis = new DataInputStream(System.in);

	// for each image...
	while (true) {
	    // read it from the stream
	    cr.read(dis);
	    
	    // "write" it to the image
	    cr.writeToImage(g);

	    // and repaint the component to view the result
	    cc.repaint();
	}
    }
    
    public static void main(String... args) throws Exception {
	new ClassViewer();
    }
}

/**
 * A simple component to render an image to.
 */
class CamComponent extends JComponent {
    BufferedImage bi = new BufferedImage(CamReader.IMG_WIDTH,
					 CamReader.IMG_HEIGHT,
					 BufferedImage.TYPE_INT_ARGB);
    
    public void paint(Graphics g1) {
	if (bi == null) return;
	Graphics2D g = (Graphics2D)g1;
	
	// scale image to fit component size
	AffineTransform aff = new AffineTransform();
	aff.scale((double)getWidth()/bi.getWidth(), 
		  (double)getHeight()/bi.getHeight());
	
	// draw
	g.drawRenderedImage(bi, aff);
    }
}

/**
 * This class reads from a specified stream into an byte array.
 * It also has a method to convert this byte array to an image.
 * 
 * 1. stream -> classified byte array
 * 2. draw classified pixel onto image
 */
class CamReader {
    final static int IMG_WIDTH = 176, IMG_HEIGHT = 144;
    
    final static Color[] col = new Color[] {
	new Color(255, 47, 16),   // ball
	new Color(79, 132, 77),   // ground
	Color.WHITE,              // wall
	new Color(255, 227, 16),  // yellow goal
	new Color(58, 94, 117),   // blue goal
	new Color(190, 93, 99),   // opponent
	Color.GRAY
    };

    // image to store the decoded chunks into (filled with classes of 
    // each pixel).
    byte[] new_img;

    CamReader() {
	new_img = new byte[IMG_WIDTH * IMG_HEIGHT];
    }

    /**
     * Reads the stream and decodes the chunks into a byte array.
     * This "image" is stored in this class in the new_img.
     *
     * Note: If we have lost track of the stream we handle the problem in 
     *       this class, finding the end of this image (4 term-chunks)..
     */
    public void read(DataInputStream dis) throws Exception {
	
	int curr_pos = 0; // the current position in the image
	byte[] chunk = new byte[2]; // a chunk consists of two bytes
	
	while(curr_pos < new_img.length) {
	    try {
		// read a chunk from the stream
		dis.readFully(chunk);
	    } catch (EOFException e) {
		break;
	    }
	    
	    // create a chunk class with the two bytes read from the stream
	    Chunk c = new Chunk(chunk[0], chunk[1]);
	    
	    // if the chunk is a term chunk then just continue
	    if(c.isTermChunk())
		continue;
	    
	    try {
		// get the class
		byte tclazz = c.getClazz();
		if(tclazz == -1) tclazz = 6;

		// and write it to the array containing the image
		Arrays.fill(new_img, curr_pos, curr_pos +c.getCount(),tclazz);
	    } catch (Exception e) {
		// if a problem occur ...
		//
		// (problem with the serial connection between the eyebot
		// and the computer, this happens!)
		//
		// ... read all chunks till the we find a the end of the image!
		
		int numTerms = 0;
		while (true) {
		    // read only one byte and check if its a term-chunk
		    numTerms = (new Chunk(dis.readByte(),(byte)0).isTermChunk()?
			numTerms+1 : 0);
		    
		    // if the number of terminals is bigger or equal to 4 we 
		    // break the loop. (We found the end of the image data!)
		    if (numTerms >= 4)
			break;
		}
		break;
	    }
	    curr_pos += c.getCount();
	}
    }

    /**
     * "Writes" the image data (new_img) information to the graphics.
     */
    public void writeToImage(Graphics g) throws IOException {
	int index = 0;
	for (int r = 0; r<IMG_HEIGHT; r++) {
	    for (int c = 0; c<IMG_WIDTH; c++) {
		// find the color representing the class.
		Color color = col[new_img[index++]];
		
		// set the color
		g.setColor(color);
		
		// draw the pixel
		g.drawLine(c,r, c,r);
	    }
	}
    }
}

/**
 * Class to represent a two byte "chunk".
 * The chunk describes the class and the number of pixels filled with
 * the specified class.
 *
 * A description of the two bytes.
 *
 *  class msb count   0   lsb count
 * |-----|---------| |-|-------------|
 *  7 6 5 4 3 2 1 0   7 6 5 4 3 2 1 0
 *
 * Note: getClazz() with zz, not to overwrite the getClass() method
 *       provided by the java vm.
 */
class Chunk {
    final int b0, b1;
    
    public Chunk(byte b0, byte b1) {
	this.b0 = b0 & 0xFF;
	this.b1 = b1 & 0xFF;
    }
    
    public int getCount() {
	int msb = (b0 & 31) << 8;
	int lsb = b1 & 255;
	return msb | lsb;
    }

    public byte getClazz() {
	return (byte)((b0 >> 5) - 1);
    }
    
    public boolean isTermChunk() {
	return getClazz() == 6;
    }
}
