/*
  See readme for program information.
  
  Compile:
    $ gcc68 -o cam.hex cam.c 
  
  Install to eyebot:
    $ dl cam.hex
*/

#include "eyebot.h"
#include <stdio.h>
#include "rgb.h" // sample RGB lut

#define BIG_WIDTH  176
#define BIG_HEIGHT 144
#define MAX_SIZE ((1<<13)-1)

typedef BYTE bigimage[BIG_HEIGHT][BIG_WIDTH][3];

// method decl
void init_cam();
void error(char str[]);
void sendRLEImgData(bigimage img);
void send(BYTE b);
void get_enc(int clazz, int count, BYTE* buf);
int  class_at(int index, bigimage img);

/*
  Program main.
 */
int main() {
  bigimage bimg;
  
  // init camera
  init_cam();
  CAMMode(AUTOBRIGHTNESS);
  CAMSet(FPS1_875, 0, 0);

  // init rs232 connection
  OSInitRS232(SER115200, NONE, SERIAL1);
  
  while (1) {
    // take a picture
    CAMGetFrameRGB((BYTE *)bimg);
    
    // compress and send it
    sendRLEImgData(bimg);
  }
    
  return 0;
}

/*
  Compresses and sends the image to the RS232 channel in chunks
  of two bytes (see get_enc(..) function).
 */
void sendRLEImgData(bigimage img)
{
  BYTE buf[2]; // the buffer to send a chunk

  int index = 0; // current index in the image to check
  int chunkCount = 0; // a sent chunk counter 
  
  // while we havn't checked the whole image
  while (index < BIG_WIDTH * BIG_HEIGHT)
    {
      // the count (that we have seen this class)
      int count = 1;
      // get the class that the specified index
      int clazz = class_at(index++, img);
      
      // as long as we see the same class (and we are not out of bounds)
      // increment the index and count variable.
      while (clazz == class_at(index, img) &&
	     count < MAX_SIZE &&
	     index < BIG_WIDTH * BIG_HEIGHT)
	{
	  index++;
	  count++;
	}
      
      // transform the class and oount variables into a two byte buffer (chunk)
      get_enc(clazz, count, buf);
      
      // send the two bytes
      send(buf[0]);
      send(buf[1]);
      
      // send a term-chunks every 100 normal chunk.
      if (chunkCount % 100 == 0) {
	send(7 << 5); // term chunk
	send(7 << 5);
      }
      
      chunkCount++;
    }

  // send two term chunks to tell the receiver that the image has been sent
  // (due to syncronization requirements)
  send(7 << 5);
  send(7 << 5);
  send(7 << 5);
  send(7 << 5);
  
  // print the number of chunks sent (debug information)
  printf("Chunk cnt:%d\n", chunkCount);
}

/*
  Compresses the class that the count variables into a two byte buffer.
  
  Detailes of the two bytes:
  
     class msb count   0   lsb count  
    |-----|---------| |-|-------------|
     7 6 5 4 3 2 1 0   7 6 5 4 3 2 1 0
*/
void get_enc(int clazz, int count, BYTE* buf)
{
  // we can only handle 7 classes
  if (clazz >= 7 || clazz < 0)
    printf("Class error.\n");
  
  // we can only handle MAX_SIZE pixels filled with the same class
  if (count > MAX_SIZE || count < 0)
    printf("Count error.\n");

  // put them into the buffer...

  clazz += 1;

  BYTE classBits = (clazz << 5) & 255;
  BYTE msbCountBits = (count >> 8) & 31;
  buf[0] = classBits | msbCountBits;
    
  BYTE lsbCountBits = count & 255;
  buf[1] = lsbCountBits;
}

/*
  This method is up to YOU to write!!!
  
  This example uses a 32k lookup table (rgb.h) to transform a RGB value
  into a class.
 */
int class_at(int index, bigimage img)
{
  int y = index / BIG_WIDTH;
  int x = index % BIG_WIDTH;
  
  BYTE r = (BYTE)img[y][x][0] >> 3;
  BYTE g = (BYTE)img[y][x][1] >> 3;
  BYTE b = (BYTE)img[y][x][2] >> 3;
  
  int ccindex = r << 10 | g << 5 | b;
  
  if (ccindex >= 32768 || ccindex < 0)
    printf("%d %d %d\n", r, g, b);
  
  return RGB_LUT[ccindex];
}

/*
  Sends a byte on the RS232 channel.
  Note: To receive all bytes you need to set the receiver in "raw" mode
        (see ClassViewer.java run instructions). If this is not the case 
	the receiver will not handle all received bytes correctly (for 
	example there is a DEL character that will delete the previous
	character in the stream).
 */
void send(BYTE b) {
  OSSendCharRS232(b, SERIAL1);
}

/*
  Initializes the camera
 */
void init_cam() {
  int camera;

  camera = CAMInit(NORMAL);

  printf("Camera type: %d\n", camera);
  
  if (camera < COLCAM) {
    error("No colour camera!");
  } else {
    if (camera == NOCAM)
      error("No camera!\n");
    else if (camera == INITERROR)
      error("CAM init error!\n");
  }
}

/*
  Error method - prints a method and exits the program
 */
void error(char str[])
{
  LCDPrintf("ERROR: %s\n", str);
  OSWait(200);
  OSExit(0);
}

